"""
Logging execution paths to make it easy to create nested paths for logging and debugging.
"""
import dataclasses
import warnings
from collections import defaultdict
from contextlib import contextmanager
from contextvars import ContextVar
from dataclasses import dataclass

from blackhc.project.utils.context_stopwatch import ContextStopwatch


@dataclass
class LogPathStep:
    step_name: str
    step_index: int
    unique_path: str
    is_summary_step: bool = False
    parent: "LogPathStep | None" = None
    used_names: dict[str, int] = dataclasses.field(
        default_factory=lambda: defaultdict(int)
    )

    def create_child(
        self, name, *, path_separator: str, is_summary_step: bool = False
    ) -> "LogPathStep":
        is_summary_step |= self.is_summary_step

        step_index = self.used_names[name]

        # Built a unique name that is fully indexed.
        if is_summary_step:
            unique_name = name

            if self.used_names[name] >= 1:
                unique_name = f"{name}+{step_index}"

                if step_index == 1:
                    # Log a warning that the name has been used before.
                    warnings.warn(
                        f"Sub name {name} has already been used in summary step. Use a unique name please!"
                    )
        else:
            unique_name = f"{name}[{step_index}]"

        # Increment the number of times the sub_name has been used.
        self.used_names[name] += 1

        if self.parent is None:
            step_name = name
            unique_path = unique_name
        else:
            step_name = f"{self.step_name}{path_separator}{name}"
            unique_path = f"{self.unique_path}{path_separator}{unique_name}"

        if is_summary_step:
            step_name = unique_path

        return LogPathStep(step_name, step_index, unique_path, is_summary_step, self)

    def metric_name(
        self, metric_name, *, path_separator: str, is_summary: bool = False
    ):
        if self.parent is None:
            return metric_name
        else:
            if is_summary:
                return f"{self.unique_path}{path_separator}{metric_name}"
            else:
                return f"{self.step_name}{path_separator}{metric_name}"


@dataclass
class LogPathNodeStats:
    count: int = 0
    first_time: float = float("nan")
    rest_time: float = float("nan")

    @property
    def mean_rest_time(self):
        if self.count == 0:
            return self.rest_time
        return self.rest_time / (self.count - 1)

    @property
    def mean_time(self):
        if self.count <= 1:
            return self.first_time
        return (self.first_time + self.rest_time) / self.count

    @property
    def total_time(self):
        if self.count <= 1:
            return self.first_time
        return self.first_time + self.rest_time

    def append_time(self, time):
        if self.count == 0:
            self.first_time = time
        else:
            if self.count == 1:
                self.rest_time = 0.0
            self.rest_time += time

        self.count += 1


class LogPath:
    """
    Log Execution Paths Manager.

    A simple class to make it easy to create nested paths for logging and debugging.
    """

    path_separator = "/"
    _ROOT: LogPathStep = LogPathStep("", 0, "")
    _current: ContextVar[LogPathStep] = ContextVar("current_xpath_node", default=_ROOT)
    all_summary_step_names = set()
    all_step_names = set()
    all_metrics = set()
    all_summary_metrics = set()
    all_step_stats = defaultdict(LogPathNodeStats)

    @classmethod
    @contextmanager
    def step(cls, name, is_summary_step: bool = False):
        """
        Create a new sub-path.

        Args:
            name: The name of the step.
            is_summary_step: If True, the step will be treated as a summary step.
                This means that a unique name will be generated for the step.
        """
        current = cls._current.get()
        child = current.create_child(
            name, path_separator=cls.path_separator, is_summary_step=is_summary_step
        )
        if child.is_summary_step:
            cls.all_summary_step_names |= {child.step_name}
        else:
            cls.all_step_names |= {child.step_name}
        token = cls._current.set(child)
        try:
            with ContextStopwatch() as stopwatch:
                yield child
            cls.all_step_stats[child.step_name].append_time(stopwatch.elapsed_time)
        finally:
            cls._current.reset(token)

    @classmethod
    @property
    def current_step_name(cls):
        """
        Get the current path.

        Returns:
            The current path.
        """
        return cls._current.get().step_name

    @classmethod
    @property
    def current_step_index(cls):
        """
        Get the current step index.

        Returns:
            The current step index.
        """
        return cls._current.get().step_index

    @classmethod
    @property
    def current_step_count(cls):
        """
        Get the current step count.

        Returns:
            The current step count.
        """
        return cls.all_step_stats[cls.current_step_name].count

    @classmethod
    @property
    def current_unique_path(cls):
        """
        Get the current unique path.

        Returns:
            The current unique path.
        """
        return cls._current.get().unique_path

    @classmethod
    def metric_name(cls, metric_name, is_summary: bool = False):
        """
        Get the full metric name.

        Args:
            metric_name:

        Returns:
            The full metric name.
        """
        metric_name = cls._current.get().metric_name(
            metric_name, path_separator=cls.path_separator, is_summary=is_summary
        )
        if is_summary:
            cls.all_summary_metrics |= {metric_name}
        else:
            cls.all_metrics |= {metric_name}
        return metric_name

    @classmethod
    def test_reset(cls):
        cls._current.set(cls._ROOT)
        cls._ROOT.used_names.clear()
        cls.all_step_names = set()
        cls.all_summary_step_names = set()
        cls.all_step_stats.clear()
        cls.all_metrics = set()
        cls.all_summary_metrics = set()


xpath = LogPath

# Limit exports
__all__ = ["xpath"]



from blackhc.project.utils.tree_namespace import TreeNamespace, items, keys, schema, values


def test_tree_namespace_simple():
    # Create a dictionary for testing
    test_dict = {"a/b": 1, "a/c": 2, "d": 3}
    # Create a DeepKeyNamespace object
    dkns = TreeNamespace(test_dict)
    # Test __getitem__ method
    assert dkns["a/b"] == 1, "Failed on __getitem__ method"
    assert dkns["a/c"] == 2, "Failed on __getitem__ method"
    assert dkns["d"] == 3, "Failed on __getitem__ method"
    # Test __len__ method
    assert len(dkns) == 2, "Failed on __len__ method"
    # Test __repr__ method
    assert (
        repr(dkns) == "TreeNamespace({'a': {'b': 1, 'c': 2}, 'd': 3})"
    ), "Failed on __repr__ method:\n" + repr(dkns)

    assert dkns.a.b == 1, "Failed on __getattribute__ method"
    assert dkns.a.c == 2, "Failed on __getattribute__ method"
    assert dkns.d == 3, "Failed on __getattribute__ method"


def test_tree_namespace_path():
    # Create a dictionary for testing
    test_dict = {"a/c": 2, "d": 3, "a/b/c": 4, "a/b/d": 5}
    # Create a DeepKeyNamespace object
    dkns = TreeNamespace(test_dict)
    # Test __getitem__ method with a sub-path
    sub_dkns = dkns["a/b"]
    assert sub_dkns["c"] == 4, "Failed on __getitem__ method"
    assert sub_dkns["d"] == 5, "Failed on __getitem__ method"
    # Test __len__ method
    assert len(sub_dkns) == 2, "Failed on __len__ method"


def test_tree_namespace_wildcard():
    # Create a dictionary for testing
    test_dict = {"a/c": 2, "d": 3, "a/b/c": 4, "a/b/d": 5}
    # Create a DeepKeyNamespace object
    dkns = TreeNamespace(test_dict)
    # Test __getitem__ method with a wildcard
    sub_dkns = dkns["a/*"]
    assert sub_dkns["c"] == 2, "Failed on __getitem__ method"
    assert sub_dkns["b/c"] == 4, "Failed on __getitem__ method"
    assert sub_dkns["b/d"] == 5, "Failed on __getitem__ method"
    # Test __len__ method
    assert len(sub_dkns) == 2, "Failed on __len__ method"


def test_tree_namespace_wildcard_2():
    # Create a dictionary for testing with two sub-arrays: a/0/0 etc
    test_dict = {"a/0/0": 1, "a/0/1": 2, "a/1/0": 3, "a/1/1": 4}
    # Create a DeepKeyNamespace object
    dkns = TreeNamespace(test_dict)
    # Test __getitem__ method with a wildcard
    sub_dkns = dkns["a/*/0"]
    assert sub_dkns["0"] == 1, "Failed on __getitem__ method"
    assert sub_dkns["1"] == 3, "Failed on __getitem__ method"
    # Test __len__ method
    assert len(sub_dkns) == 2, "Failed on __len__ method"

    # Test a double wildcard
    sub_dkns = dkns["a/*/*"]
    assert sub_dkns["0/0"] == 1, "Failed on __getitem__ method"
    assert sub_dkns["0/1"] == 2, "Failed on __getitem__ method"
    assert sub_dkns["1/0"] == 3, "Failed on __getitem__ method"
    assert sub_dkns["1/1"] == 4, "Failed on __getitem__ method"
    # Test __len__ method
    assert len(sub_dkns) == 2, "Failed on __len__ method"
    # Test sub access
    assert sub_dkns[0][0] == 1, "Failed on __getitem__ method"
    assert sub_dkns[0][1] == 2, "Failed on __getitem__ method"
    assert sub_dkns[1][0] == 3, "Failed on __getitem__ method"
    assert sub_dkns[1][1] == 4, "Failed on __getitem__ method"


def test_tree_namespace_wildcard_at():
    # Create a dictionary for testing with a nested key of interest (loss) at different nesting levels
    test_dict = {
        "a/0/0/loss": 1,
        "a/0/1/loss": 2,
        "a/1/0/loss": 3,
        "a/1/1/loss": 4,
        "a/loss": 5,
    }
    # Create a DeepKeyNamespace object
    dkns = TreeNamespace(test_dict)
    # Test __getitem__ method with a wildcard
    sub_dkns = dkns["@/loss"]
    assert sub_dkns["a_0_0"] == 1, "Failed on __getitem__ method"
    assert sub_dkns["a_0_1"] == 2, "Failed on __getitem__ method"
    assert sub_dkns["a_1_0"] == 3, "Failed on __getitem__ method"
    assert sub_dkns["a_1_1"] == 4, "Failed on __getitem__ method"
    assert sub_dkns["a"] == 5, "Failed on __getitem__ method"
    # Test __len__ method
    assert len(sub_dkns) == 5, "Failed on __len__ method"


def test_tree_namespace_keys():
    # Create a dictionary for testing
    test_dict = {"a/b": 1, "a/c": 2, "d": 3}
    # Create a DeepKeyNamespace object
    dkns = TreeNamespace(test_dict)
    # Test keys method
    assert keys(dkns) == ["a", "d"], "Failed on keys method"


def test_tree_namespace_values():
    # Create a dictionary for testing
    test_dict = {"a/b": 1, "a/c": 2, "d": 3}
    # Create a DeepKeyNamespace object
    dkns = TreeNamespace(test_dict)
    # Test values method
    assert values(dkns) == [dkns.a, dkns.d], f"Failed on values method:\n{values(dkns)}"


def test_tree_namespace_items():
    # Create a dictionary for testing
    test_dict = {"a/b": 1, "a/c": 2, "d": 3}
    # Create a DeepKeyNamespace object
    dkns = TreeNamespace(test_dict)
    # Test items method
    assert items(dkns) == [("a", dkns.a), ("d", dkns.d)], "Failed on items method"


def test_tree_namespace_schema():
    # Create a dictionary for testing
    test_dict = {"a/b": 1, "a/c": 2, "d": 3}
    # Create a DeepKeyNamespace object
    dkns = TreeNamespace(test_dict)
    # Test schema method
    assert schema(dkns) == {"a/b": int, "a/c": int, "d": int}, "Failed on schema method"
    assert schema(dkns) == {
        "a": {"b": int, "c": int},
        "d": int,
    }, "Failed on schema method"


if __name__ == "__main__":
    test_tree_namespace_simple()
    test_tree_namespace_path()
    test_tree_namespace_wildcard()
    test_tree_namespace_wildcard_2()
    test_tree_namespace_wildcard_at()
    test_tree_namespace_keys()
    test_tree_namespace_values()
    test_tree_namespace_items()
    test_tree_namespace_schema()



"""
A simple decorator around dataclass that makes it easier to create config classes.
"""
import dataclasses
import functools

__all__ = ["configclass"]

import inspect


def __init_config_class(self, /, **kwargs):
    """
    Initialize the config class.

    Either use the keyword arguments to set the fields, or use the default values.

    Args:
        self: The config class.
        **kwargs: The keyword arguments to set the fields.

    Returns:
        None
    """
    fields = dataclasses.fields(self)
    # Check kwargs only contains fields
    for key in kwargs:
        if key not in [field.name for field in fields]:
            raise ValueError(f"Unknown field {key} for {self.__class__}")

    for field in fields:
        if field.name not in kwargs:
            # Get the class attribute (if it exists).
            class_default = getattr(self.__class__, field.name, None)
            if class_default is not None:
                # Check if it is a Field (factory) or a value.
                if isinstance(class_default, dataclasses.Field):
                    # Check if the field has a default factory.
                    if class_default.default_factory is not dataclasses.MISSING:
                        # Call the default factory.
                        kwargs[field.name] = class_default.default_factory()
                    # Check if the field has a default value.
                    elif class_default.default is not dataclasses.MISSING:
                        # Use the default value.
                        kwargs[field.name] = class_default.default
                    else:
                        raise ValueError(
                            f"Field {field.name} does not have a default value for {self.__class__}"
                        )
                else:
                    # Use the class attribute.
                    kwargs[field.name] = class_default
            else:
                raise ValueError(
                    f"Field {field.name} for {self.__class__} has no default value"
                )

    self.__dict__.update(kwargs)


# Custom __iter__ to allow for unpacking.
def __iter_config_class(self):
    """Iterate over the fields (their values) of the config class.

    The field values are returned in the order they are defined in the class.

    Args:
        self:

    Returns:
        An iterator over the fields of the config class.
    """
    for field in dataclasses.fields(self):
        yield getattr(self, field.name)


# Custom mapping methods to allow for dict-like access:
# __getitem__, __len__, __contains__, keys, items, values, get
def __getitem_config_class(self, key):
    """Get the value of a field in the config class."""
    return getattr(self, key)


def __len_config_class(self):
    """Get the number of fields in the config class."""
    return len(dataclasses.fields(self))


def __contains_config_class(self, key):
    """Check if the config class contains a field."""
    return hasattr(self, key)


def __keys_config_class(self):
    """Get the keys of the config class."""
    return [field.name for field in dataclasses.fields(self)]


def __items_config_class(self):
    """Get the items of the config class."""
    return [
        (field.name, getattr(self, field.name)) for field in dataclasses.fields(self)
    ]


def __values_config_class(self):
    """Get the values of the config class."""
    return [getattr(self, field.name) for field in dataclasses.fields(self)]


# Add | operator to allow for merging config classes (strict).
def __or_config_class(self, other):
    """Merge two config classes."""
    # Verify that self can contain all fields from `other`.
    other_dict = {**other}
    # Check that all fields in other are in self.
    missing_fields = set(other_dict.keys()) - set(self.keys())
    if missing_fields:
        raise ValueError(
            f"Cannot merge {other} into {self} because {type(other)} contains fields that are not in {type(self)}: {missing_fields}"
        )

    merged_fields = {**self, **other}
    return self.__class__(**merged_fields)


def __invoke_config_class(self, func):
    """Invoke a callable with config class's fields."""
    # Assert that func is callable.
    if not callable(func):
        raise ValueError(f"{func} is not callable")
    # Filter the fields of the config class using the callable's signature.
    signature = inspect.signature(func)
    # Get the parameters of the callable.
    parameters = signature.parameters
    # Get the fields of the config class.
    kwargs = {**self}
    # Filter the fields of the config class using the parameters of the callable.
    # (If there are variadic keyword arguments, then do not filter.)
    if all(
        parameter.kind != inspect.Parameter.VAR_KEYWORD
        for parameter in parameters.values()
    ):
        kwargs = {
            key: kwargs[key]
            for key in parameters
            if key in kwargs and parameters[key].kind != inspect.Parameter.VAR_KEYWORD
        }
    # Invoke the callable with the filtered fields.
    return func(**kwargs)


# Create a configclass decorator that wraps dataclass
@functools.wraps(dataclasses.dataclass)
def configclass(cls, /, **kwargs):
    """Create a config class from a dataclass.

    Args:
        cls:
        **kwargs:

    Returns:

    """
    # Assert that there is no init keyword argument.
    if "init" in kwargs:
        raise ValueError("Cannot specify init keyword argument for configclass")
    # Assert no eq keyword argument.
    if "eq" in kwargs:
        raise ValueError("Cannot specify eq keyword argument for configclass")
    # Assert that there is no __init__ method.
    custom_methods = {
        "__init__": __init_config_class,
        "__iter__": __iter_config_class,
        "__getitem__": __getitem_config_class,
        "__len__": __len_config_class,
        "__contains__": __contains_config_class,
        "keys": __keys_config_class,
        "items": __items_config_class,
        "values": __values_config_class,
        "invoke": __invoke_config_class,
        "__or__": __or_config_class,
    }
    # Check if the class has any of the custom methods.
    for method_name, method in custom_methods.items():
        if method_name in cls.__dict__:
            raise ValueError(
                f"Cannot specify {method_name} method for configclass ({cls} already has this method)!"
            )
    # Set the custom methods.
    for method_name, method in custom_methods.items():
        setattr(cls, method_name, method)
    # Create the dataclass with the class and the arguments.
    return dataclasses.dataclass(init=False, eq=True, **kwargs)(cls)


configclass.__doc__ = """Returns the same class as was passed in, with dunder methods
added based on the fields defined in the class. This is a wrapper around dataclass.

It uses the class variables to set the default values for the fields---so changing
class variables will change the (future) default values for the fields.

Further, on instantiation of a configclass instance, the fields can be set using
keyword arguments. If a field is not set, the default value is used.

Two configclasses can be merged using the | operator. The fields of the right-hand
side configclass are used to overwrite the fields of the left-hand side configclass.

configclasses can be iterated over, and they can be accessed and unpacked like a dict.

It supports fields with default values, fields with default factories, and fields
with no default value.

It adds the following methods to the class:
    __init__(self, /, **kwargs)
    __iter__(self)
    __getitem__(self, key)
    __len__(self)
    __contains__(self, key)
    keys(self)
    items(self)
    values(self)
    __or__(self, other)

It also always adds __eq__, and __ne__ methods to the class via dataclass.

Examines PEP 526 __annotations__ to determine fields.

If repr is true, a __repr__() method is added. If order is true, rich
comparison dunder methods are added. If unsafe_hash is true, a
__hash__() method function is added. If frozen is true, fields may
not be assigned to after instance creation. If match_args is true,
the __match_args__ tuple is added. If kw_only is true, then by
default all fields are keyword-only. If slots is true, an
__slots__ attribute is added.
"""


# Test the configclass decorator.
if __name__ == "__main__":
    # Create a config class.
    @configclass
    class Config:
        a: int
        b: int
        c: int = 3

    # Create an instance of the config class.
    config = Config(a=1, b=2)
    # Print the config class.
    print(config)
    # Print the config class as a dict.
    print(dict(config))
    # Print the config class as a list.
    print(list(config))
    # Print the config class as a tuple.
    print(tuple(config))
    # Print the config class as a set.
    print(set(config))
    # Print the config class as a list of keys.
    print(list(config.keys()))
    # Print the config class as a list of values.
    print(list(config.values()))
    # Print the config class as a list of items.
    print(list(config.items()))

    # Change the default value of the config class.
    Config.c = 4
    # Create an instance of the config class.
    config = Config(a=1, b=2)
    # Print the config class.
    print(config)



"""
Component class to make it easy to create component based classes.

(This can allow for more robust code that can be iterated a lot.)
"""
import typing
from dataclasses import dataclass

T = typing.TypeVar("T")
C = typing.TypeVar("C", bound="Component")


def explicit_try_cast(cls: typing.Type[T], instance) -> T | None:
    """Try to cast an instance to a protocol.

    If the instance implements the protocol, return the instance. Otherwise, return None.
    """
    if isinstance(instance, cls):
        return instance
    elif isinstance(instance, Component):
        return instance.query_protocol(cls)
    else:
        return None


def explicit_cast(cls: typing.Type[T], instance) -> T:
    """Cast an instance to a protocol.

    If the instance implements the protocol, return the instance. Otherwise, raise a TypeError.
    """
    view = explicit_try_cast(cls, instance)
    if view is None:
        raise TypeError(f"Cannot cast {type(instance)} to {cls} for:\n{instance}")
    return view


@typing.runtime_checkable
class Interface(typing.Protocol):
    """
    Interface for a component.

    This is a protocol that can be used to check if a class implements an interface and to cast to an interface.
    """

    @classmethod
    def try_cast(cls: typing.Type[T], instance) -> T | None:
        """Try to cast an instance to the class.

        If the instance implements the class, return the instance. Otherwise, return None.
        """
        return explicit_try_cast(cls, instance)

    @classmethod
    def cast(cls: typing.Type[T], instance) -> T:
        """Cast an instance to the class.

        If the instance implements the class, return the instance. Otherwise, raise a TypeError.
        """
        return explicit_cast(cls, instance)


@typing.runtime_checkable
class Component(Interface, typing.Protocol):
    """
    Default component implementation.
    """

    def query_protocol(self, cls: typing.Type[T]) -> T | None:
        """Query the protocol for a component.

        If the component implements the protocol, return the component. Otherwise, return None.
        """
        if isinstance(self, cls):
            return self
        return None


@dataclass
class ComponentView(Component, typing.Generic[C]):
    """
    Adapter for an interface.

    When you want to implement a different interface for a component, you can use this class.

    Example
    -------

    >>> class InterfaceA(Interface):
    ...     def a(self):
    ...         raise NotImplementedError
    >>> class InterfaceB(Interface):
    ...     def a(self):
    ...         raise NotImplementedError
    >>> class ImplementationAB(InterfaceA, Component):
    ...     def a(self):
    ...         return 0
    ...     def b(self):
    ...         return 1
    ...     def query_protocol(self, cls: typing.Type[T]) -> T:
    ...         if issubclass(InterfaceB, cls):
    ...             return ImplementationAB.ViewB(self)
    ...     class ViewB(InterfaceB, ComponentView['ImplementationAB']):
    ...         def a(self):
    ...             return self._component.b()
    """

    _component: C

    def query_protocol(self, cls: typing.Type[T]) -> T | None:
        """Query the protocol for a component.

        If the component implements the protocol, return the component. Otherwise, return None.
        """
        if isinstance(self, cls):
            return self

        return self._component.query_protocol(cls)


@dataclass
class ProtocolWrapper(Component, typing.Generic[T]):
    """Adapter for a protocol.

    When you want to implement a different protocol for a component, you can use this class.
    """

    protocol_type: typing.Type[T]
    instance: T

    def query_protocol(self, cls: typing.Type[T]) -> T | None:
        if issubclass(cls, self.protocol_type):
            return self.instance
        return super().query_protocol(cls)



"""
Component tests.
"""
from blackhc.project.utils.component import *


class InterfaceA(Interface):
    def a(self):
        raise NotImplementedError


class InterfaceB(Interface):
    def b(self):
        raise NotImplementedError


@typing.runtime_checkable
class ProtocolC(typing.Protocol):
    def c(self):
        raise NotImplementedError


class ImplementationAB(InterfaceA, InterfaceB):
    def a(self):
        return 0

    def b(self):
        return 1

    def c(self):
        return 2


def test_interface_ab():
    implementation_ab = ImplementationAB()
    assert implementation_ab.a() == 0
    assert implementation_ab.b() == 1

    assert isinstance(implementation_ab, ImplementationAB)
    assert isinstance(implementation_ab, InterfaceA)
    assert isinstance(implementation_ab, InterfaceB)
    assert isinstance(implementation_ab, Interface)

    assert InterfaceA.cast(implementation_ab) is implementation_ab
    assert InterfaceB.cast(implementation_ab) is implementation_ab
    assert isinstance(implementation_ab, ProtocolC)

    assert ImplementationAB.cast(implementation_ab) is implementation_ab


class ImplementationA(InterfaceA):
    def a(self):
        return 0


def test_interface_a():
    implementation_a = ImplementationA()

    assert not isinstance(implementation_a, InterfaceB)


@dataclass
class ImplementationB(InterfaceB, Component):
    field: int

    def b(self):
        return self.field

    def query_protocol(self, cls: typing.Type[T]) -> T:
        if issubclass(InterfaceA, cls):
            return ImplementationB_ViewA(self)
        return super().query_protocol(cls)


class ImplementationB_ViewA(InterfaceA, ComponentView[ImplementationB]):
    def a(self):
        return self._component.field


def test_component():
    implementation_b = ImplementationB(1)
    interface_a = InterfaceA.cast(implementation_b)
    assert interface_a.a() == 1
    interface_b = InterfaceB.cast(interface_a)
    assert interface_b.b() == 1
    assert isinstance(interface_b, InterfaceB)



import operator

from blackhc.project.utils.early_stopping import EarlyStopping


def test_early_stopping_improves():
    # Test case where score improves every time
    early_stopping = EarlyStopping(patience=3, lower_is_better=True)
    scores = [0.5, 0.4, 0.3, 0.2, 0.1]
    for score in scores:
        assert early_stopping.step(score) > 0
    assert early_stopping.best_score == 0.1
    assert early_stopping.cmp == operator.lt


def test_early_stopping_improves_with_higher_is_better():
    # Test case where score improves every time and higher is better
    early_stopping = EarlyStopping(patience=3, lower_is_better=False)
    scores = [0.1, 0.2, 0.3, 0.4, 0.5]
    for score in scores:
        assert early_stopping.step(score) > 0
    assert early_stopping.best_score == 0.5
    assert early_stopping.cmp == operator.gt


def test_early_stopping_does_not_improve():
    import pytest

    # Test case where score never improves
    early_stopping = EarlyStopping(patience=2, lower_is_better=True)
    scores = [0.5, 0.6, 0.7, 0.8, 0.9]
    for i, score in enumerate(scores):
        if i < 3:
            assert early_stopping.step(score)
        elif i == 3:
            assert not early_stopping.step(score)
        else:
            with pytest.raises(AssertionError):
                early_stopping.step(score)
    assert early_stopping.best_score == 0.5
    assert early_stopping.cmp == operator.lt


def test_early_stopping_improves_sometimes():
    # Test case where score improves sometimes
    early_stopping = EarlyStopping(patience=1, lower_is_better=True)
    scores = [0.5, 0.6, 0.4, 0.7, 0.3]
    for i, score in enumerate(scores):
        assert early_stopping.step(score)
        assert early_stopping.patience != (i % 2)
    assert early_stopping.best_score == 0.3
    assert early_stopping.cmp == operator.lt


def test_early_stopping_iter():
    # Test case for __iter__ method
    early_stopping = EarlyStopping(patience=1, lower_is_better=True)
    scores = [0.5, 0.6, 0.7, 0.8]
    for score, _ in zip(scores, early_stopping):
        early_stopping.step(score)
    assert early_stopping.best_score == 0.5
    assert early_stopping.cmp == operator.lt
    assert score == 0.7


def test_early_stopping_to_str():
    # Test case for __str__ method
    early_stopping = EarlyStopping(patience=1, lower_is_better=True)
    assert str(early_stopping) == "EarlyStopping(patience=1, lower_is_better=True)"



import pytest
import torch
import platform

from blackhc.project.utils import cpu_memory


def is_apple_silicon():
    return platform.machine() == 'arm64'


@pytest.mark.skipif(is_apple_silicon(), reason="Apple Silicon")
@pytest.mark.forked
def test_cpu_mem_limit():
    # 128 MB (128/4M float32)
    tensor = torch.empty((128, 1024, 1024 // 4), dtype=torch.float32)
    tensor.resize_(1)

    cpu_memory.set_cpu_memory_limit(0.25)

    # 512 MB (128/4M float32)
    with pytest.raises(RuntimeError):
        torch.empty((512, 1024, 1024 // 4), dtype=torch.float32)



from pyfakefs import fake_filesystem
from blackhc import project
import os


def test_get_cookiecutter_project_path_from_notebooks(fs: fake_filesystem.FakeFilesystem):
    fs.CreateDirectory('/tmp/blackhc.project/notebooks')
    assert (project.get_cookiecutter_project_path('/tmp/blackhc.project/notebooks') == os.path.abspath(
        '/tmp/blackhc.project'))


def test_get_cookiecutter_project_path_from_scripts(fs: fake_filesystem.FakeFilesystem):
    fs.CreateDirectory('/tmp/blackhc.project/scripts')
    assert (project.get_cookiecutter_project_path('/tmp/blackhc.project/notebooks') == os.path.abspath(
        '/tmp/blackhc.project'))


def test_get_cookiecutter_project_path_with_src(fs: fake_filesystem.FakeFilesystem):
    fs.CreateDirectory('/tmp/blackhc.project/src')
    assert (
            project.get_cookiecutter_project_path('/tmp/blackhc.project/') == os.path.abspath('/tmp/blackhc.project'))


def test_get_cookiecutter_project_path_without_src(fs: fake_filesystem.FakeFilesystem):
    fs.CreateDirectory('/tmp/blackhc.project')
    assert project.get_cookiecutter_project_path('/tmp/blackhc.project/') is None



import functools
from contextlib import contextmanager

try:
    import wandb
except ImportError:
    raise ImportError("Package 'wandb' is not installed. Please install it using 'pip install wandb'")

from blackhc.project.utils.log_path import xpath


@contextmanager
def wandb_custom_step(name):
    """
    A context manager for wandb steps.


    Args:
        name: The name of the step.
        is_summary_step: Whether this is a summary step.

    Returns:
        The context manager.
    """

    with xpath.step(name, is_summary_step=False):
        step_metric = xpath.metric_name("step", is_summary=False)
        wandb.log({step_metric: xpath.current_step_index}, commit=False)
        if xpath.current_step_count == 0:
            wandb.define_metric(f"{xpath.current_step_name}/*", step_metric)
        try:
            yield
        finally:
            commit()


def log_metric(metric_name, value, is_summary=False):
    """
    Log a metric to wandb.

    Args:
        metric_name: The metric name.
        value: The metric value.
        is_summary: Whether this is a summary metric.
    """
    metric_name = xpath.metric_name(metric_name, is_summary=is_summary)
    if is_summary is False:
        wandb.log({metric_name: value}, commit=False)
    else:
        wandb.run.summary[metric_name] = value


def log(metrics, is_summary=False):
    """
    Log metrics to wandb.

    Args:
        metrics: The metrics.
        is_summary: Whether this is a summary metric.
    """
    metrics = {
        xpath.metric_name(k, is_summary=is_summary): v for k, v in metrics.items()
    }
    if is_summary is False:
        return wandb.log(metrics, commit=False)
    else:
        return wandb.run.summary.update(metrics)


def commit():
    """
    Commit the metrics to wandb.
    """
    return wandb.log({}, commit=True)


@functools.wraps(wandb.define_metric)
def define_metric(metric_name, *args, **kwargs):
    """
    Define a metric for wandb.

    Args:
        metric_name: The metric name.
        ...: The other arguments.
    """
    metric_name = xpath.metric_name(metric_name)
    return wandb.define_metric(metric_name, *args, **kwargs)



# Tests for config_class.
import pytest

from blackhc.project.utils.config_class import configclass


@configclass
class TestConfigClass:
    a: int
    b: str
    c: float


def test_config_class():
    config = TestConfigClass(a=1, b="2", c=3.0)
    assert config.a == 1
    assert config.b == "2"
    assert config.c == 3.0

    assert config["a"] == 1
    assert config["b"] == "2"
    assert config["c"] == 3.0


def test_config_class_conversions():
    config = TestConfigClass(a=1, b="2", c=3.0)

    assert dict(config) == {"a": 1, "b": "2", "c": 3.0}
    assert list(config) == [1, "2", 3.0]
    assert tuple(config) == (1, "2", 3.0)
    assert set(config) == {1, "2", 3.0}
    assert list(config.keys()) == ["a", "b", "c"]
    assert list(config.values()) == [1, "2", 3.0]
    assert list(config.items()) == [("a", 1), ("b", "2"), ("c", 3.0)]


def test_config_class_default_value():
    # throws exception if we don't pass c
    with pytest.raises(ValueError):
        TestConfigClass(a=1, b="2")

    TestConfigClass.c = 4.0

    config = TestConfigClass(a=1, b="2")

    assert config.c == 4.0


def test_config_class_merge_via_or():
    config = TestConfigClass(a=1, b="2", c=3.0)
    config2 = TestConfigClass(a=2, b="3", c=4.0)

    config3 = config | config2

    assert config3.a == 2
    assert config3.b == "3"
    assert config3.c == 4.0

    # test merging with a dict
    config4 = config | {"a": 3, "b": "4", "c": 5.0}

    assert config4.a == 3
    assert config4.b == "4"
    assert config4.c == 5.0


def test_config_class_unpacking():
    config = TestConfigClass(a=1, b="2", c=3.0)

    a, b, c = config

    assert a == 1
    assert b == "2"
    assert c == 3.0

    assert {**config} == {"a": 1, "b": "2", "c": 3.0}


def test_config_class_invoke():
    @configclass
    class TestConfigClass:
        a: int
        b: str
        c: float

    def func(a, b, c):
        return a, b, c

    config = TestConfigClass(a=1, b="2", c=3.0)
    result = config.invoke(func)
    assert result == (1, "2", 3.0)


def test_config_class_invoke_kwargs():
    @configclass
    class TestConfigClass:
        a: int
        b: str
        c: float

    def func(**kwargs):
        return kwargs["a"], kwargs["b"], kwargs["c"]

    config = TestConfigClass(a=1, b="2", c=3.0)
    result = config.invoke(func)
    assert result == (1, "2", 3.0)
